#include "CLanguageAddOn.h"
#include <stack>
#include <ctype.h>

extern "C" {
_EXPORT const char kLanguageName[] = "Zadig's Reveng";
_EXPORT const char kLanguageExtensions[] = "rev;";
_EXPORT const char kLanguageCommentStart[] = "--";
_EXPORT const char kLanguageCommentEnd[] = "";
_EXPORT const char kLanguageKeywordFile[] = "keywords.rev";
}

enum {
	START, 
	IDENT, 
	STRING,  // "strings"
	LCOMMENT,
	NUMERIC,
	SYMBOL,
};

#define GETCHAR (c = (i++ < size) ? text[i - 1] : 0)

bool isSymbol(char c)
{
	if (c == '(' || c == ')' || c == '[' || c == ']' || c == '&' || c == '%' ||
		c == ',' || c == ';' || c == '$' || c == '#' || c == '*')
		return true;
	
	return false;
}

bool isNumeric(char c)
{
	if (c >= '0' && c <= '9')
		return true;

	return false;
}

bool isHexNum(char c)
{
	if ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'))
		return true;

	return false;
}

_EXPORT void ColorLine(CLanguageProxy& proxy, int& state)
{
	const char *text = proxy.Text();
	int size = proxy.Size();
	int i = 0, s = 0, kws = 0, esc = 0;
	char c;
	bool leave = false;
	bool hex_num = false;

	if (state == LCOMMENT)
		proxy.SetColor(0, kLCommentColor);
	else
		proxy.SetColor(0, kLTextColor);
	
	if (size <= 0)
		return;
	
	while (!leave)
	{
		GETCHAR;
		
		switch (state) {
			case START:

				if (isHexNum(c) && isHexNum(text[i]) && (isHexNum(text[i+1]) ||
					isHexNum(text[i-1] || text[i+1] == ':')))
				{
					kws = proxy.Move(c, 1);
					state = NUMERIC;
				}
				else if (isalpha(c) || c == '_' || c == '.')
				{
					kws = proxy.Move(c, 1);
					state = IDENT;
				}
				else if (c == '/' && text[i] == '/')
				{
					i++;
					state = LCOMMENT;
				}
				else if (c == '"')
					state = STRING;
				else if (isNumeric(c) || (c == '$' && isHexNum(text[i])))
				{
					state = NUMERIC;
				}
				else if (isSymbol(c))
				{
					state = SYMBOL;
				}
				else if (c == '\n' || c == 0)
					leave = true;
					
				if (leave || (state != START && s < i))
				{
					proxy.SetColor(s, kLTextColor);
					s = i - 1;
				}
			break;
			
			// // format comments
			case LCOMMENT:
				proxy.SetColor(s - 1, kLCommentColor);
				leave = true;
				if (text[size - 1] == '\n')
					state = START;
			break;

			case IDENT:
				if (!isalnum(c) && c != '_')
				{
					int kwc;

					if (i > s + 1 && (kwc = proxy.IsKeyWord(kws)) != 0)
					{
						switch (kwc)
						{
							case 1:	proxy.SetColor(s, kLKeyWordColor); break;
							case 2:	proxy.SetColor(s, kLUser1); break;
							case 3:	proxy.SetColor(s, kLUser2); break;
							case 4:	proxy.SetColor(s, kLUser3); break;
							case 5:	proxy.SetColor(s, kLUser4); break;
						}
						s = --i;
						state = START;
					}
					else
					{
						proxy.SetColor(s, kLTextColor);
						s = --i;
						state = START;
					}
				}
				else if (kws)
					kws = proxy.Move((int)(unsigned char) c, kws);
			break;
			
			case STRING:
				if (c == '"' && !esc)
				{
					proxy.SetColor(s, kLStringColor);
					s = i;
					state = START;
				}
				else if (c == '\n' || c == 0)
				{
					if (text[i - 2] == '\\' && text[i - 3] != '\\')
					{
						proxy.SetColor(s, kLStringColor);
					}
					else
					{
						proxy.SetColor(s, kLTextColor);
						state = START;
					}
					
					s = size;
					leave = true;
				}
				else
					esc = !esc && (c == '\\');
			break;

			case NUMERIC:
				proxy.SetColor(s, kLNumberColor);
				if (isNumeric(text[i - 1]) || isHexNum(text[i - 1]))
					;
				else
					if (text[i - 1] == 'x' && hex_num == false)
						hex_num = true;
					else
					{
						s = i - 1;
						i--;
						hex_num = false;
						state = START;
					}
			break;
		
			case SYMBOL:
				proxy.SetColor(s, kLSeparatorColor);
				if (isSymbol(text[i - 1]))
					;
				else
				{
					s = i - 1;
					i--;
					state = START;
				}
			break;			
			
			default:
				leave = true;
			break;
		}
	}
}